# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from dyff.schema.platform import Entities
from dyff.storage.config import config


def auditprocedure_notebook(auditprocedure_id: str) -> str:
    return f"{config.resources.auditprocedures.storage.url}/{auditprocedure_id}/notebook.ipynb"


def auditreport_html(audit_id: str) -> str:
    return f"{auditreport_root(audit_id)}/index.html"


def auditreport_root(audit_id: str) -> str:
    return f"{config.resources.auditreports.storage.url}/{audit_id}"


def dataset_root(dataset_id: str) -> str:
    return f"{config.resources.datasets.storage.url}/{dataset_id}"


def dataset_data(dataset_id: str) -> str:
    return f"{dataset_root(dataset_id)}/data"


def dataset_strata(dataset_id: str) -> str:
    # FIXME: Handle multi-file data (or choose a better name than 'part-0')
    return f"{dataset_root(dataset_id)}/strata/part-0.parquet"


def dataset_task(dataset_id: str, task_id: str) -> str:
    return f"{config.resources.datasets.storage.url}/{dataset_id}/tasks/{task_id}"


def datasource_root(datasource_id: str) -> str:
    return f"{config.resources.datasources.storage.url}/{datasource_id}"


def measurements_root(measurement_id: str) -> str:
    return f"{config.resources.measurements.storage.url}/{measurement_id}"


def module_root(module_id: str) -> str:
    return f"{config.resources.modules.storage.url}/{module_id}"


def outputs_raw(evaluation_id: str) -> str:
    return f"{config.resources.outputs.storage.url}/{evaluation_id}/data"


def outputs_verified(evaluation_id: str) -> str:
    return f"{config.resources.outputs.storage.url}/{evaluation_id}/verified"


def report_data(report_id: str) -> str:
    # FIXME: Handle multi-file data (or choose a better name than 'part-0')
    return f"{report_root(report_id)}/part-0.parquet"


def report_root(report_id: str) -> str:
    return f"{config.resources.reports.storage.url}/{report_id}"


def safetycase_root(safetycase_id: str) -> str:
    return f"{config.resources.safetycases.storage.url}/{safetycase_id}"


def inferenceservice_source_archive(inferenceservice_id: str) -> str:
    return f"{config.resources.inferenceservices.storage.url}/{inferenceservice_id}/source.tar.gz"


def model_source_archive(model_id: str) -> str:
    return f"{config.resources.models.storage.url}/{model_id}/source.tar.gz"


def logs_file_for_entity(entity_root: str) -> str:
    return f"{entity_root}/.dyff/logs.txt"


def entity_artifacts_root(kind: Entities, id: str) -> str:
    if kind == Entities.Dataset:
        return dataset_root(id)
    elif kind == Entities.Evaluation:
        return outputs_verified(id)
    elif kind == Entities.Measurement:
        return measurements_root(id)
    elif kind == Entities.Module:
        return module_root(id)
    elif kind == Entities.Report:
        return report_root(id)
    elif kind == Entities.SafetyCase:
        return safetycase_root(id)
    else:
        raise ValueError(f"entity kind {kind} has no associated artifacts")



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import abc
from typing import Collection, NamedTuple, Optional

from dyff.schema.platform import (
    Audit,
    Dataset,
    DataSource,
    Documentation,
    Evaluation,
    Family,
    InferenceService,
    InferenceSession,
    Measurement,
    Method,
    Model,
    Module,
    Report,
    SafetyCase,
    Score,
)
from dyff.schema.requests import (
    AuditQueryRequest,
    DatasetQueryRequest,
    DocumentationEditRequest,
    DocumentationQueryRequest,
    EvaluationQueryRequest,
    FamilyQueryRequest,
    InferenceServiceQueryRequest,
    InferenceSessionQueryRequest,
    MeasurementQueryRequest,
    MethodQueryRequest,
    ModelQueryRequest,
    ModuleQueryRequest,
    ReportQueryRequest,
    SafetyCaseQueryRequest,
    ScoreQueryRequest,
)


class Whitelist(NamedTuple):
    accounts: set[str]
    entities: set[str]

    @staticmethod
    def everything() -> Whitelist:
        return Whitelist(accounts=set(["*"]), entities=set(["*"]))

    @staticmethod
    def nothing() -> Whitelist:
        return Whitelist(accounts=set(), entities=set())

    @staticmethod
    def _intersection_star(a: set[str], b: set[str]) -> set[str]:
        if "*" in a:
            return b
        elif "*" in b:
            return a
        else:
            return a.intersection(b)

    def intersection(self, other: Whitelist) -> Whitelist:
        return Whitelist(
            accounts=Whitelist._intersection_star(self.accounts, other.accounts),
            entities=Whitelist._intersection_star(self.entities, other.entities),
        )


class QueryBackend(abc.ABC):
    @abc.abstractmethod
    def get_audit(self, id: str) -> Optional[Audit]:
        """Retrieve an Audit entity.

        Parameters:
          id: The unique key of the Audit.

        Returns:
          The Audit, or None if no Audit with the specified key exists.
        """

    @abc.abstractmethod
    def query_audits(
        self, whitelist: Whitelist, query: AuditQueryRequest
    ) -> Collection[Audit]:
        """Retrieve all Audit entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Audit entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_data_source(self, id: str) -> Optional[DataSource]:
        """Retrieve a DataSource entity.

        Parameters:
          id: The unique key of the DataSource.

        Returns:
          The DataSource, or None if no DataSource with the specified key exists.
        """

    @abc.abstractmethod
    def query_data_sources(
        self, whitelist: Whitelist, **query
    ) -> Collection[DataSource]:
        """Retrieve all DataSource entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the DataSource entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_dataset(self, id: str) -> Optional[Dataset]:
        """Retrieve a Dataset entity.

        Parameters:
          id: The unique key of the Dataset.

        Returns:
          The Dataset, or None if no Dataset with the specified key exists.
        """

    @abc.abstractmethod
    def query_datasets(
        self, whitelist: Whitelist, query: DatasetQueryRequest
    ) -> Collection[Dataset]:
        """Retrieve all Dataset entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Dataset entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def edit_documentation(
        self, id: str, edit: DocumentationEditRequest
    ) -> Optional[Documentation]:
        """Edit the Documentation entity associated with a resource.

        Parameters:
          id: ID of the documented resource.
          edit: Edit request containing changes to make to the documentation.

        Returns:
          The new Documentation, or None if no resource exists with the given ID.
        """

    @abc.abstractmethod
    def get_documentation(self, id: str) -> Optional[Documentation]:
        """Retrieve the Documentation entity associated with a resource.

        Parameters:
          id: ID of the documented resource.

        Returns:
          The Documentation, or None if no resource exists with the given ID.
        """

    @abc.abstractmethod
    def query_documentation(
        self, whitelist: Whitelist, query: DocumentationQueryRequest
    ) -> Collection[Documentation]:
        """Retrieve all Documentation entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: The query request.
        """

    @abc.abstractmethod
    def get_evaluation(self, id: str) -> Optional[Evaluation]:
        """Retrieve an Evaluation entity.

        Parameters:
          id: The unique key of the Evaluation.

        Returns:
          The Evaluation, or None if no Evaluation with the specified key exists.
        """

    @abc.abstractmethod
    def query_evaluations(
        self, whitelist: Whitelist, query: EvaluationQueryRequest
    ) -> Collection[Evaluation]:
        """Retrieve all Evaluation entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Evaluation entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_family(self, id: str) -> Optional[Family]:
        """Retrieve a Family entity.

        Parameters:
          id: The unique ID of the Family.

        Returns:
          The Family, or None if no Family with the specified ID exists.
        """

    @abc.abstractmethod
    def query_families(
        self, whitelist: Whitelist, query: FamilyQueryRequest
    ) -> Collection[Family]:
        """Retrieve all Family entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Family entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """

    @abc.abstractmethod
    def get_inference_service(self, id: str) -> Optional[InferenceService]:
        """Retrieve an InferenceService entity.

        Parameters:
          id: The unique key of the InferenceService.

        Returns:
          The InferenceService, or None if no InferenceService with the specified key exists.
        """

    @abc.abstractmethod
    def query_inference_services(
        self, whitelist: Whitelist, query: InferenceServiceQueryRequest
    ) -> Collection[InferenceService]:
        """Retrieve all InferenceService entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the InferenceService entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_inference_session(self, id: str) -> Optional[InferenceSession]:
        """Retrieve an InferenceSession entity.

        Parameters:
          id: The unique key of the InferenceSession.

        Returns:
          The InferenceSession, or None if no InferenceSession with the specified key exists.
        """

    @abc.abstractmethod
    def query_inference_sessions(
        self, whitelist: Whitelist, query: InferenceSessionQueryRequest
    ) -> Collection[InferenceSession]:
        """Retrieve all InferenceSession entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the InferenceSession entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_measurement(self, id: str) -> Optional[Measurement]:
        """Retrieve a Measurement entity.

        Parameters:
          id: The unique key of the Measurement.

        Returns:
          The Measurement, or None if no Measurement with the specified key exists.
        """

    @abc.abstractmethod
    def query_measurements(
        self, whitelist: Whitelist, query: MeasurementQueryRequest
    ) -> Collection[Measurement]:
        """Retrieve all Measurement entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Measurement entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_method(self, id: str) -> Optional[Method]:
        """Retrieve a Method entity.

        Parameters:
          id: The unique key of the Method.

        Returns:
          The Method, or None if no Method with the specified key exists.
        """

    @abc.abstractmethod
    def query_methods(
        self, whitelist: Whitelist, query: MethodQueryRequest
    ) -> Collection[Method]:
        """Retrieve all Method entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Method entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_model(self, id: str) -> Optional[Model]:
        """Retrieve a Model entity.

        Parameters:
          id: The unique key of the Model.

        Returns:
          The Model, or None if no Model with the specified key exists.
        """

    @abc.abstractmethod
    def query_models(
        self, whitelist: Whitelist, query: ModelQueryRequest
    ) -> Collection[Model]:
        """Retrieve all Model entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Model entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_module(self, id: str) -> Optional[Module]:
        """Retrieve a Module entity.

        Parameters:
          id: The unique key of the Module.

        Returns:
          The Module, or None if no Module with the specified key exists.
        """

    @abc.abstractmethod
    def query_modules(
        self, whitelist: Whitelist, query: ModuleQueryRequest
    ) -> Collection[Module]:
        """Retrieve all Module entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Module entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_report(self, id: str) -> Optional[Report]:
        """Retrieve a Report entity.

        Parameters:
          id: The unique key of the Report.

        Returns:
          The Report, or None if no Report with the specified key exists.
        """

    @abc.abstractmethod
    def query_reports(
        self, whitelist: Whitelist, query: ReportQueryRequest
    ) -> Collection[Report]:
        """Retrieve all Report entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the Report entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_safetycase(self, id: str) -> Optional[SafetyCase]:
        """Retrieve a SafetyCase entity.

        Parameters:
          id: The unique key of the SafetyCase.

        Returns:
          The SafetyCase, or None if no SafetyCase with the specified key exists.
        """

    @abc.abstractmethod
    def query_safetycases(
        self, whitelist: Whitelist, query: SafetyCaseQueryRequest
    ) -> Collection[SafetyCase]:
        """Retrieve all SafetyCase entities matching the query parameters.

        Parameters:
          whitelist: The set of accounts and entities that the caller has
            been granted access to.
          **query: Equality constraints on fields of the SafetyCase entity.
            The returned entities satisfy 'entity.field==value' for all items
            'field: value' in kwargs.
        """

    @abc.abstractmethod
    def get_score(self, id: str) -> Optional[Score]:
        """Retrieve a Score entity.

        :param id: The unique key of the Score.
        :returns: The Score, or None if no Score with the specified key exists.
        """

    @abc.abstractmethod
    def query_scores(
        self, whitelist: Whitelist, query: ScoreQueryRequest
    ) -> Collection[Score]:
        """Retrieve all Score entities matching the query parameters.

        :param whitelist: The set of accounts and entities that the caller has been
            granted access to.
        :param query: Equality constraints on fields of the Score entity. The returned
            entities satisfy 'entity.field==value' for all items 'field: value' in
            kwargs.
        """



# SPDX-FileCopyrightText: 2024 UL Research Institutes
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Optional

from dyff.schema.platform import (
    Audit,
    Dataset,
    DataSource,
    EntityStatus,
    EntityStatusReason,
    Evaluation,
    Family,
    FamilyMembers,
    InferenceService,
    InferenceSession,
    Labeled,
    Measurement,
    Method,
    Model,
    Module,
    Report,
    SafetyCase,
)


class CommandBackend(abc.ABC):
    @abc.abstractmethod
    def update_status(
        self, id: str, *, status: str, reason: Optional[str] = None
    ) -> None:
        """Update the status of an entity.

        :param id: The entity .id
        :type id: str
        :param status: New .status value
        :type status: str
        :param reason: New .reason value
        :type reason: Optional[str]
        """

    @abc.abstractmethod
    def update_labels(self, id: str, labels: Labeled) -> None:
        """Updated the labels of a labeled entity.

        :param id: The ID of the entity to update.
        :type id: str
        :param labels: The labels to update.
        :type labels: Labeled
        """

    @abc.abstractmethod
    def update_family_members(self, id: str, members: FamilyMembers) -> None:
        """Updated the members of a Family.

        :param id: The ID of the Family to update.
        :type id: str
        :param members: The members to update.
        :type members: FamilyMembers
        """

    @abc.abstractmethod
    def create_audit(self, spec: Audit) -> Audit:
        """Create a new Audit entity in the system.

        :param spec: Specification of the Audit. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: Audit
        :return: A copy of ``spec`` with all system fields set.
        :rtype: Audit
        """

    @abc.abstractmethod
    def create_data_source(self, spec: DataSource) -> DataSource:
        """Create a new DataSource entity in the system.

        :param spec: Specification of the DataSource. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: DataSource
        :return: A copy of ``spec`` with all system fields set.
        :rtype: DataSource
        """

    @abc.abstractmethod
    def create_dataset(self, spec: Dataset) -> Dataset:
        """Create a new Dataset entity in the system.

        :param spec: Specification of the Dataset. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: Dataset
        :return: A copy of ``spec`` with all system fields set.
        :rtype: Dataset
        """

    @abc.abstractmethod
    def create_evaluation(self, spec: Evaluation) -> Evaluation:
        """Create a new Evaluation entity in the system.

        :param spec: Specification of the Evaluation. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: Evaluation
        :return: A copy of ``spec`` with all system fields set.
        :rtype: Evaluation
        """

    @abc.abstractmethod
    def create_family(self, spec: Family) -> Family:
        """Create a new Family entity in the system.

        :param spec: Specification of the Family. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: Family
        :return: A copy of ``spec`` with all system fields set.
        :rtype: Family
        """

    @abc.abstractmethod
    def create_inference_service(self, spec: InferenceService) -> InferenceService:
        """Create a new InferenceService entity in the system.

        :param spec: Specification of the InferenceService. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: InferenceService
        :return: A copy of ``spec`` with all system fields set.
        :rtype: InferenceService
        """

    @abc.abstractmethod
    def create_inference_session(self, spec: InferenceSession) -> InferenceSession:
        """Create a new InferenceSession entity in the system.

        :param spec: Specification of the InferenceSession. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: InferenceSession
        :return: A copy of ``spec`` with all system fields set.
        :rtype: InferenceSession
        """

    @abc.abstractmethod
    def create_measurement(self, spec: Measurement) -> Measurement:
        """Create a new Measurement entity in the system.

        :param spec: Specification of the Measurement. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: Measurement
        :return: A copy of ``spec`` with all system fields set.
        :rtype: Measurement
        """

    @abc.abstractmethod
    def create_method(self, spec: Method) -> Method:
        """Create a new Method entity in the system.

        :param spec: Specification of the Method. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: Method
        :return: A copy of ``spec`` with all system fields set.
        :rtype: Method
        """

    @abc.abstractmethod
    def create_model(self, spec: Model) -> Model:
        """Create a new Model entity in the system.

        :param spec: Specification of the Model. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: Model
        :return: A copy of ``spec`` with all system fields set.
        :rtype: Model
        """

    @abc.abstractmethod
    def create_module(self, spec: Module) -> Module:
        """Create a new Module entity in the system.

        :param spec: Specification of the Module. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: Module
        :return: A copy of ``spec`` with all system fields set.
        :rtype: Module
        """

    @abc.abstractmethod
    def create_report(self, spec: Report) -> Report:
        """Create a new Report entity in the system.

        :param spec: Specification of the Report. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: Report
        :return: A copy of ``spec`` with all system fields set.
        :rtype: Report
        """

    @abc.abstractmethod
    def create_safetycase(self, spec: SafetyCase) -> SafetyCase:
        """Create a new SafetyCase entity in the system.

        :param spec: Specification of the SafetyCase. The system fields of the spec
            such as ``.id`` must be **unset**.
        :type spec: SafetyCase
        :return: A copy of ``spec`` with all system fields set.
        :rtype: SafetyCase
        """

    def terminate_workflow(self, id: str) -> None:
        """Terminate a running workflow.

        :param id: The ID of the workflow.
        :type id: str
        """
        self.update_status(
            id,
            status=EntityStatus.terminated,
            reason=EntityStatusReason.terminate_command,
        )

    def delete_entity(self, id: str) -> None:
        """Delete an existing entity.

        :param id: The ID of the entity.
        :type id: str
        """
        self.update_status(
            id,
            status=EntityStatus.deleted,
            reason=EntityStatusReason.delete_command,
        )
